@@ -77,10 +77,7 @@ static int send_ping(struct backend_logind *backend) {
 	int ret = sd_bus_call_method_async(backend->bus, NULL, "org.freedesktop.login1",
 					   "/org/freedesktop/login1", "org.freedesktop.DBus.Peer",
 					   "Ping", ping_handler, backend, "");
-	if (ret < 0) {
-		return ret;
-	}
-	return 0;
+	return ret < 0 ? ret : 0;
 }
 
 static void check_pending_events(struct backend_logind *backend) {
@@ -122,7 +119,6 @@ static int open_device(struct libseat *base, const char *path, int *fd) {
 				 major(st.st_rdev), minor(st.st_rdev));
 	if (ret < 0) {
 		log_errorf("Could not take device: %s", error.message);
-		tmpfd = -1;
 		goto out;
 	}
 
@@ -130,7 +126,6 @@ static int open_device(struct libseat *base, const char *path, int *fd) {
 	ret = sd_bus_message_read(msg, "hb", &tmpfd, &paused);
 	if (ret < 0) {
 		log_errorf("Could not parse D-Bus response: %s", strerror(-ret));
-		tmpfd = -1;
 		goto out;
 	}
 
@@ -138,8 +133,8 @@ static int open_device(struct libseat *base, const char *path, int *fd) {
 	// so we just clone it.
 	tmpfd = fcntl(tmpfd, F_DUPFD_CLOEXEC, 0);
 	if (tmpfd < 0) {
-		log_errorf("Could not duplicate fd: %s", strerror(errno));
-		tmpfd = -1;
+		ret = -errno;
+		log_errorf("Could not duplicate fd: %s", strerror(-ret));
 		goto out;
 	}
 
@@ -149,6 +144,10 @@ out:
 	sd_bus_error_free(&error);
 	sd_bus_message_unref(msg);
 	check_pending_events(session);
+	if (ret < 0) {
+		errno = -ret;
+		return -1;
+	}
 	return tmpfd;
 }
 
@@ -179,7 +178,11 @@ static int close_device(struct libseat *base, int device_id) {
 	sd_bus_error_free(&error);
 	sd_bus_message_unref(msg);
 	check_pending_events(session);
-	return ret < 0 ? -1 : 0;
+	if (ret < 0) {
+		errno = -ret;
+		return -1;
+	}
+	return 0;
 }
 
 static int switch_session(struct libseat *base, int s) {
@@ -202,7 +205,11 @@ static int switch_session(struct libseat *base, int s) {
 	sd_bus_error_free(&error);
 	sd_bus_message_unref(msg);
 	check_pending_events(session);
-	return ret < 0 ? -1 : 0;
+	if (ret < 0) {
+		errno = -ret;
+		return -1;
+	}
+	return 0;
 }
 
 static int disable_seat(struct libseat *base) {
@@ -388,7 +395,7 @@ static int handle_pause_device(sd_bus_message *msg, void *userdata, sd_bus_error
 		set_active(session, false);
 		ret = sd_bus_call_method(session->bus, "org.freedesktop.login1", session->path,
 					 "org.freedesktop.login1.Session", "PauseDeviceComplete",
-					 ret_error, &msg, "uu", major, minor);
+					 ret_error, NULL, "uu", major, minor);
 		if (ret < 0) {
 			log_errorf("Could not send PauseDeviceComplete signal: %s",
 				   ret_error->message);
@@ -428,7 +435,6 @@ static int handle_properties_changed(sd_bus_message *msg, void *userdata, sd_bus
 			goto error;
 		}
 		if (strcmp(s, "Active") == 0) {
-			int ret;
 			const char *field = "Active";
 			ret = sd_bus_message_enter_container(msg, 'v', "b");
 			if (ret < 0) {
@@ -555,6 +561,9 @@ static int manager_get_session_path(struct backend_logind *session, char **sessi
 	}
 	free(*session_path);
 	*session_path = strdup(path);
+	if (*session_path == NULL) {
+		ret = -ENOMEM;
+	}
 
 out:
 	sd_bus_error_free(&error);
@@ -584,6 +593,9 @@ static int manager_get_seat_path(struct backend_logind *session, char **seat_pat
 	}
 	free(*seat_path);
 	*seat_path = strdup(path);
+	if (*seat_path == NULL) {
+		ret = -ENOMEM;
+	}
 
 out:
 	sd_bus_error_free(&error);
@@ -605,6 +617,10 @@ static int get_display_session(char **session_id) {
 			goto error;
 		}
 		*session_id = strdup(xdg_session_id);
+		if (*session_id == NULL) {
+			ret = -ENOMEM;
+			goto error;
+		}
 		goto success;
 	}
 
