@@ -141,7 +141,7 @@ static int vt_ack(struct seat *seat, bool release) {
  * 3. Otherwise, the first client on the seat's list of clients, if any.
  *
  * Be careful not to call seat_activate immediately after closing a client, as
- * this can lead to it immediatley re-opening. The client should be removed as
+ * this can lead to it immediately re-opening. The client should be removed as
  * a candidate before seat_activate is called.
  */
 static int seat_activate(struct seat *seat) {
@@ -182,7 +182,7 @@ done:
 }
 
 /*
- * seat_add_client assigns a sesssion ID to the client and adds it to the seat,
+ * seat_add_client assigns a session ID to the client and adds it to the seat,
  * if allowed. The client does not open the seat, remaining closed until
  * `seat_open_client` is called.
  *
@@ -229,7 +229,16 @@ int seat_add_client(struct seat *seat, struct client *client) {
 		}
 		client->session = seat->cur_vt;
 	} else {
-		client->session = seat->session_cnt++;
+		int next_session = 1;
+		for (struct linked_list *elem = seat->clients.next; elem != &seat->clients;
+		     elem = elem->next) {
+			struct client *c = (struct client *)elem;
+			if (c->session == next_session) {
+				next_session++;
+				elem = &seat->clients;
+			}
+		}
+		client->session = next_session;
 	}
 
 	client->seat = seat;
@@ -259,9 +268,7 @@ void seat_remove_client(struct client *client) {
 
 	while (!linked_list_empty(&client->devices)) {
 		struct seat_device *device = (struct seat_device *)client->devices.next;
-		if (seat_close_device(client, device) == -1) {
-			log_errorf("Could not close %s: %s", device->path, strerror(errno));
-		}
+		seat_close_device(client, device);
 	}
 
 	bool was_current = seat->active_client == client;
@@ -324,7 +331,7 @@ struct seat_device *seat_open_device(struct client *client, const char *path) {
 	log_debugf("Opening device %s for client %d on %s", path, client->session, seat->seat_name);
 
 	if (client->state != CLIENT_ACTIVE) {
-		log_error("Could open device: client is not active");
+		log_error("Could not open device: client is not active");
 		errno = EPERM;
 		return NULL;
 	}
@@ -353,14 +360,13 @@ struct seat_device *seat_open_device(struct client *client, const char *path) {
 
 	int device_id = 1;
 	size_t device_count = 0;
-	struct seat_device *device = NULL;
 	for (struct linked_list *elem = client->devices.next; elem != &client->devices;
 	     elem = elem->next) {
 		struct seat_device *old_device = (struct seat_device *)elem;
 
 		if (strcmp(old_device->path, sanitized_path) == 0) {
 			old_device->ref_cnt++;
-			return device;
+			return old_device;
 		}
 
 		if (old_device->device_id >= device_id) {
@@ -401,7 +407,7 @@ struct seat_device *seat_open_device(struct client *client, const char *path) {
 		abort();
 	}
 
-	device = calloc(1, sizeof(struct seat_device));
+	struct seat_device *device = calloc(1, sizeof(struct seat_device));
 	if (device == NULL) {
 		log_errorf("Allocation failed: %s", strerror(errno));
 		close(fd);
@@ -414,6 +420,7 @@ struct seat_device *seat_open_device(struct client *client, const char *path) {
 		log_errorf("Allocation failed: %s", strerror(errno));
 		close(fd);
 		free(device);
+		errno = ENOMEM;
 		return NULL;
 	}
 
@@ -470,15 +477,20 @@ static int seat_deactivate_device(struct seat_device *seat_device) {
  * seat_close_device reduces the reference count for the device. If it reaches
  * zero, the device is deactivated, closed and removed.
  */
-int seat_close_device(struct client *client, struct seat_device *seat_device) {
+void seat_close_device(struct client *client, struct seat_device *seat_device) {
 	log_debugf("Closing device %s for client %d on %s", seat_device->path, client->session,
 		   client->seat->seat_name);
 
 	seat_device->ref_cnt--;
 	if (seat_device->ref_cnt > 0) {
-		return 0;
+		return;
 	}
 
+	// The caller might be closing devices in error handling. As we cannot
+	// fail anyway, let's ensure we do not clobber errno for caller
+	// convenience.
+	int stored_errno = errno;
+
 	linked_list_remove(&seat_device->link);
 	if (seat_device->fd != -1) {
 		seat_deactivate_device(seat_device);
@@ -486,7 +498,8 @@ int seat_close_device(struct client *client, struct seat_device *seat_device) {
 	}
 	free(seat_device->path);
 	free(seat_device);
-	return 0;
+
+	errno = stored_errno;
 }
 
 /*
@@ -542,7 +555,7 @@ int seat_open_client(struct seat *seat, struct client *client) {
 
 	if (seat->vt_bound && vt_open(client->session) == -1) {
 		log_error("Could not open VT for client");
-		goto error;
+		return -1;
 	}
 
 	for (struct linked_list *elem = client->devices.next; elem != &client->devices;
@@ -553,21 +566,23 @@ int seat_open_client(struct seat *seat, struct client *client) {
 		}
 	}
 
-	client->state = CLIENT_ACTIVE;
-	seat->active_client = client;
 	if (client_send_enable_seat(client) == -1) {
 		log_error("Could not send enable signal to client");
-		goto error;
+		for (struct linked_list *elem = client->devices.next; elem != &client->devices;
+		     elem = elem->next) {
+			struct seat_device *device = (struct seat_device *)elem;
+			seat_deactivate_device(device);
+		}
+		if (seat->vt_bound) {
+			vt_close(client->session);
+		}
+		return -1;
 	}
 
+	client->state = CLIENT_ACTIVE;
+	seat->active_client = client;
 	log_infof("Opened client %d on %s", client->session, seat->seat_name);
 	return 0;
-
-error:
-	if (seat->vt_bound) {
-		vt_close(seat->cur_vt);
-	}
-	return -1;
 }
 
 /*
@@ -639,7 +654,7 @@ int seat_ack_disable_client(struct client *client) {
 
 /*
  * seat_set_next_session queues a new client to be opened based on its session
- * ID. It can only b eperformed by an active client, and only if a switch has
+ * ID. It can only be performed by an active client, and only if a switch has
  * not already been requested. If the seat is VT-bound, a VT switch is
  * performed and the VT ack/release mechanism takes care of the rest to avoid
  * conflicts between the two mechanisms.
